/* Non-linear Ordinal Logistic Regression

   Mark Mathieson   mathies@stats.ox.ac.uk
   University of Oxford, UK.
   Based on the network code of B.D.Ripley

   Version October 1996
*/

#include <S.h>
#ifdef WIN32
#  include <newredef.h>
#endif
#include <stdio.h>
#include <math.h>

#define N 1000
#define NWTS 5000
#define NCLASS 20
#define PHIMAR 1.0e10
#define SIGTOL 250.0

#define MINPRB 1e-300

/*#define max(a,b) a>b?a:b
#define min(a,b) a<b?a:b*/
#define true ((!false))
#define false 0

#define GRADSTOP 1e-3
#define STARTSTEP 1.0

#define SCALEa 1.0
#define SCALEb 1.0

#define stepredn	0.2
#define acctol		1.0e-5
#define reltest		10.0
#define abstol		1.0e-4
#define reltol		1.0e-8
#define REPORT		10

typedef unsigned char boolean;

boolean         notcomp;
boolean         quiet = false;
boolean         nolr = !false;

static int      Epoch;
static float    Decay[NWTS];
static float    Alpha;
static float    Lambda;
static double   TotalError;

static int	Nunits; 
static int	Ninputs;
static int	FirstHidden;
static int	FirstOutput;
static int	Noutputs;
static int	NSunits;
static int	Nweights;
static int	Linout;
static int      Skip;
static int      Riskfit;
static int      CutReg;

static float	Outputs[N];
static float	ErrorSums[N];
static float	Errors[N];
static int	Nconn[N];
static int	Conn[NWTS];
static float	wts[NWTS];
static float	Slopes[NWTS];
static float    ClassSlopes[NCLASS];
static float    EtaSlopes[NWTS];
static float	Probs[N];
static float    Phi[NCLASS];
static float    Loss[(NCLASS*NCLASS)];

static int      K;
static int	NTrain;
static int	NTest;
static float	**TrainIn;
static float	**TrainOut;
static float	**TestIn;
static float	**TestOut;
static float	*Weights;
static float    toutputs[N];
static float    *h, **w;
static float    **H;

/* Curvature penalty variables */
static float    Mu;
static float    Gini;
static float    doutdx[N];
static float    doutdx2[N];
static float    delta[N];
static float    fac[N];
static float    dedw[NWTS];
static float    detadxdw[NWTS];
static int      **wlook;

static float    DGini;
static float    t1weights[NWTS];
static float    t1cutpoints[NCLASS];
static float	*DG;

static double slog(x)
double x;
{
if(x < 1.e-250)return (double)(-576.);
else return log(x);
}

static void errmsg(string)
     char *string;
{
  quiet = false;
  PROBLEM "%s\n", string RECOVER(NULL_ENTRY);
}

static void errmsg_unix(string)
     char *string;
{
  fprintf(stderr, "%s\n", string);
  exit(1);
}

static float *vect(n)
     int n;
{
  float *v;
  v = Calloc(n, float);
  if (!v) errmsg("allocation failure in vect()");
  return v;
}

static int *ivect(n)
     int n;
{
  int *v;
  v = Calloc(n, int);
  if (!v) errmsg("allocation failure in vect()");
  return v;
}

static void free_vect(v)
     float *v;
{
  Free(v);
}

/* float matrix */

static float **matrix(nrh,nch)
     int nrh,nch;
{
  int i;
  float **m;
  
  m = Calloc((nrh+1), float*);
  if (!m) errmsg("allocation failure 1 in matrix()");
  
  for(i = 0;i <= nrh;i++) {
    m[i] = Calloc((nch+1), float);
    if (!m[i]) errmsg("allocation failure 2 in matrix()");
  }
  return m;
}

static void free_matrix(m,nrh,nch)
     float **m;
     int nrh,nch;
{
  int i;
  
  for(i = nrh;i >= 0;i--) Free(m[i]);
  Free(m);
}

static float **Lmatrix(n)
     int n;
{
  int i;
  float **m;
  
  m = Calloc(n, float*);
  if (!m) errmsg("fail1 in Lmatrix()");
  
  for(i = 0;i < n;i++) {
    m[i] = Calloc((i+1), float);
    if (!m[i]) errmsg("fail2 in Lmatrix()");
  }
  return m;
}

static void free_Lmatrix(m, n)
     float **m;
     int n;
{
  int i;
  
  for (i = n - 1; i >= 0; i--) Free(m[i]);
  Free(m);
}

/* integer matrix */

static int **imatrix(nrh,nch)
     int nrh,nch;
{
  int i;
  int **m;
  
  m = Calloc((nrh+1), int*);
  if (!m) errmsg("allocation failure 1 in imatrix()");
  
  for(i = 0;i <= nrh;i++) {
    m[i] = Calloc((nch+1), int);
    if (!m[i]) errmsg("allocation failure 2 in imatrix()");
  }
  return m;
}

static void free_imatrix(m,nrh,nch)
     int **m;
     int nrh,nch;
{
  int i;
  
  for(i = nrh;i >= 0;i--) Free(m[i]);
  Free(m);
}


static Build_Net(ninputs, nhidden, noutputs)
     int  ninputs, nhidden, noutputs;
{
  int  i;

  Nunits = 1 + ninputs + nhidden + noutputs;
  if (Nunits > N) {
    fprintf(stderr, "%d is too many units\n", Nunits);
    exit(1);
  }
  Ninputs = ninputs;
  FirstHidden = 1 + ninputs;
  FirstOutput = 1 + ninputs + nhidden;
  Noutputs = noutputs;

  for (i = 0; i < Nunits; i++) {
    Outputs[i] = 0.0;
    Nconn[i] = 0;
  }
  Nconn[Nunits] = 0;
  Outputs[0] = 1.0;
}


void set_nolr(n, nconn, conn, decay, nsunits, skip, k, riskfit, loss,
	     cutreg,lambda,alpha,mu,gini,dgini)
     long *n, *nconn, *conn, *nsunits;
     float *decay,*loss,*lambda,*alpha,*mu,*gini,*dgini;
     int *k,*skip,*riskfit,*cutreg;
{
  int i,j;

  if (sizeof(TrainIn)>10)
    free_matrix(TrainIn,NTrain,Ninputs);
  if (sizeof(TrainOut)>10)
    free_matrix(TrainOut,NTrain,(Noutputs+1));

  Skip=*skip;
  K=*k;
  Build_Net((int)n[0], (int)n[1], (int)n[2]);
  for (i = 0; i <= Nunits; i++) Nconn[i] = nconn[i];
  Nweights = Nconn[Nunits];
  for (i = 0; i < Nweights; i++) {
    Conn[i] = conn[i];
    Decay[i] = decay[i];
  }
  Epoch = 0;
  Riskfit = *riskfit;
  if(Riskfit) for (i=0; i<(K*K); i++)
    Loss[i]=loss[i];

  CutReg = *cutreg;
  Lambda = *lambda;
  Mu = *mu;
  Gini = *gini;
  DGini = *dgini;
  Alpha = *alpha;
  TotalError = 0.0;
  NSunits = *nsunits;
  Linout = (NSunits < Nunits);

  if ((Mu>0) || (Gini>0) || (DGini>0)) {
    
    if (sizeof(wlook)>4)
      free_imatrix(wlook,Nunits,Nunits);
    
    wlook = imatrix(Nunits,Nunits);
	
    for (j = FirstHidden; j < Nunits; j++) 
      for (i = Nconn[j]; i < Nconn[j+1]; i++) {
	wlook[Conn[i]][j] = i;
      }    
  }
}

void set_otrain(ntr, train, weights, dg)
     long *ntr;
     float *train, *weights, *dg;
{
  int i,j;

  if (sizeof(Weights)>10)
    free_vect(Weights);
  if (sizeof(DG)>10)
    free_vect(DG);
  
  NTrain = *ntr;
  TrainIn = matrix(NTrain,Ninputs);
  TrainOut = matrix(NTrain,(Noutputs+1));
  Weights = vect(NTrain);
  DG = vect(NTrain);
  for(j=0; j < Ninputs; j++)
    for(i=0; i < NTrain; i++)
      TrainIn[i][j] = *train++; 
  for(j=0; j < (Noutputs+1); j++)
    for(i=0; i< NTrain; i++) 
      TrainOut[i][j] = *train++;
  for(i=0; i < NTrain; i++) {
    Weights[i] = *weights++;
    DG[i] = *dg++;
  }

}

void onntest(ntest, test, result)
     long *ntest;
     float *test, *result;
{
  int i,j;

  NTest = *ntest;
  if (Nweights==0)
	errmsg("No model set");
  TestIn = matrix(NTest,Ninputs);
  TestOut = matrix(NTest,Noutputs);
  for (i = 0; i < Ninputs; i++)
    for(j = 0; j < NTest; j++)
      TestIn[j][i] = *test++;

  quiet = true;
  for (i=0; i<Noutputs; i++)
    toutputs[i] = 1;
  for (j=0; j<NTest; j++){
    fpass(TestIn[j], toutputs, 1.0);
    for (i=0; i<Noutputs; i++)
      TestOut[j][i] = Outputs[FirstOutput+i];
  }
  quiet = false;

  for (i=0; i<Noutputs; i++)
    for(j=0; j<NTest; j++)
      *result++ = TestOut[j][i];
  free_matrix(TestIn,NTest,Ninputs);
  free_matrix(TestOut,NTest,Noutputs);
}

void set_parameters(inwts,phi)
     float *inwts, *phi;
{
  int i;

  for (i=1; i<K; i++)
    Phi[i]=phi[i-1];
  Phi[0]=-PHIMAR; Phi[K]=PHIMAR;
  
  for (i=0; i< Nweights; i++)
    wts[i] = inwts[i];
}


static float L(a,b)
     int a,b;
{
  float l;

  l = Loss[(b-1)*K+(a-1)];
  return(l);
}

static double sigmoid(sum)
     double sum;
{
  if (sum < -SIGTOL) 
    return(0.0);
  else if (sum > SIGTOL) 
    return(1.0);
  else 
    return (1.0 /(1.0 + exp(-sum)));
}

static double sigmoid_prime(value)
     double value;
{
  return (value * (1.0 - value));
}

static double sigmoid_prime_prime(value)
     double value;
{
  return (value * (1.0 - value) * (1.0 - 2.0*value));
}

static double sigmoid_prime_prime_prime(value)
     double value;
{
  return(value * (1.0 - value) * (1.0 + 6.0 * value * value - 6.0 * value));
}

static double prob(eta,class,classb)
     double eta;
     int class,classb;
{
  return(sigmoid(Phi[class] - eta) - sigmoid(Phi[classb] - eta));
}

static double prob_e(eta,class,classb)
     double eta;
     int class,classb;
{
  return(-sigmoid_prime(sigmoid(Phi[class] - eta))
	 + sigmoid_prime(sigmoid(Phi[classb] - eta)));
}

static double prob_p(eta,class,classb,s)
     double eta;
     int class,classb,s;
{
  if (class==s)
    return(sigmoid_prime(sigmoid(Phi[class] - eta)));
  else if (classb==s)
    return(-1.0 * sigmoid_prime(sigmoid(Phi[classb] - eta)));
  else
    return(0.0);
}

static double prob_p_p(eta,class,classb,s,t)
     double eta;
     int class,classb,s,t;
{
  if ((class==s) && (class==t))
    return(sigmoid_prime_prime(sigmoid(Phi[class] - eta)));
  else if ((classb==s) && (classb==t))
    return(-1.0 * sigmoid_prime_prime(sigmoid(Phi[classb] - eta)));
  else
    return(0.0);
}

static double prob_e_e(eta,class,classb)
     double eta;
     int class,classb;
{
  return(sigmoid_prime_prime(sigmoid(Phi[class] - eta))
	 - sigmoid_prime_prime(sigmoid(Phi[classb] - eta)));
}

/* d prob(class|eta) / (d eta d eta d eta) */
static double prob_e_e_e(eta,class,classb)
     double eta;
     int class,classb;
{
  return(-1.0 * sigmoid_prime_prime_prime(sigmoid(Phi[class] - eta))
	 + sigmoid_prime_prime_prime(sigmoid(Phi[classb] - eta)));
}

static double prob_p_e(eta,class,classb,s)
     double eta;
     int class,classb,s;
{
  return(-prob_p_p(eta,class,classb,s,s));
}

static double prob_p_e_e(eta,class,classb,s)
     double eta;
     int class,s;
{
  if (class==s)
    return(sigmoid_prime_prime_prime(sigmoid(Phi[class] - eta)));
  else if (classb==s)
    return(-1.0 * sigmoid_prime_prime_prime(sigmoid(Phi[classb] - eta)));
  else
    return(0.0);
}

static double prob_t(eta,class,classb,s)
     double eta;
     int class,s;
{
  float tmp;
  tmp = 0.0;

  if (s<=class)
    tmp += sigmoid_prime(sigmoid(Phi[class] - eta));
  if (s<=classb)
    tmp += -1.0 * sigmoid_prime(sigmoid(Phi[classb] - eta));

  return(tmp);
}

static double prob_t_e(eta,class,classb,s)
     double eta;
     int class,s;
{
  float tmp;
  tmp = 0.0;

  if (s<=class)
    tmp += -1.0 * sigmoid_prime_prime(sigmoid(Phi[class] - eta));
  if (s<=classb)
    tmp += sigmoid_prime_prime(sigmoid(Phi[classb] - eta));

  return(tmp);
}

static double prob_t_e_e(eta,class,classb,s)
     double eta;
     int class,s;
{
  float tmp;
  tmp = 0.0;

  if (s<=class)
    tmp += sigmoid_prime_prime_prime(sigmoid(Phi[class] - eta));
  if (s<=classb)
    tmp += -1.0 * sigmoid_prime_prime_prime(sigmoid(Phi[classb] - eta));

  return(tmp);
}


static fpass(input, goal, wx)
     float *input, *goal, wx;
{
  int i, j;
  double sum, t;
  float loss, prb;

  for (i = 0; i < Ninputs; i++)
    Outputs[i+1] = input[i];
  
  for (j = FirstHidden; j < Nunits; j++) {
      sum = 0.0;
      for (i = Nconn[j]; i < Nconn[j+1]; i++)
	sum += Outputs[Conn[i]] * wts[i];
      if(j < NSunits) sum = sigmoid(sum);
      Outputs[j] = sum;
    }
  
  for (i = FirstOutput; i < Nunits; i++) {
    if(Riskfit) {
      for (j=1; j<=K; j++) {
	loss = L((int)goal[0],j);
	TotalError += wx * loss * prob(Outputs[FirstOutput],j,(j-1));
      }
    } else {
      prb = prob(Outputs[FirstOutput],(int)goal[0],(int)goal[1]);
      if (prb < MINPRB)
	if (quiet) {
	  notcomp = true;
	  prb = 1.0;
	}
	else {
	  errmsg("Cannot evaluate log-likelihood");
	}
      TotalError += -1.0 * wx * slog(prb);
    }
  }
  if (abs(Outputs[FirstOutput])>(PHIMAR-5*SIGTOL))
    errmsg("Eta too large for quick calculations of p(class|eta)");

}

static bpass (goal, wx)
     float *goal, wx;
{
  int i, j, cix, k;
  float sum, loss, dltpi;
  double tmp, tmp2;

  if(Riskfit) {
    for (j=1; j<=K; j++) {
      dltpi = 0.0;
      loss = L((int)goal[0],j);
      for (i=0; i<(K-1); i++)
	ClassSlopes[i] += wx * loss * 
	  prob_p(Outputs[FirstOutput],j,(j-1),(i+1));
    }
  } else {
    tmp = prob(Outputs[FirstOutput], (int)goal[0], (int)goal[1]);
    ClassSlopes[((int)goal[0])-1] += -1.0 * wx
      * prob_p(Outputs[FirstOutput],(int)goal[0],(int)goal[1],(int)goal[0])
      / tmp;
    ClassSlopes[((int)goal[1])-1] += -1.0 * wx
      * prob_p(Outputs[FirstOutput],(int)goal[0],(int)goal[1],(int)goal[1])
      / tmp;
  }
  
  for (i = 0; i < Nweights; i++)
    EtaSlopes[i] = 0.0;

  ErrorSums[FirstOutput] = 1.0;
  for (i = FirstHidden; i < FirstOutput; i++)
    ErrorSums[i] = 0.0;
  
  for (j = Nunits-1; j >= FirstHidden; j--) {
    Errors[j] = ErrorSums[j];
    if (j < FirstOutput) Errors[j] *= sigmoid_prime(Outputs[j]); 
    for (i = Nconn[j]; i < Nconn[j+1]; i++) {
      cix = Conn[i];
      ErrorSums[cix] += Errors[j] * wts[i];
      EtaSlopes[i] += wx * Errors[j] * Outputs[cix];
    }
  }
  
  if(Riskfit) {
    for (j=1; j<=K; j++) {
      loss = L((int)goal[0],j);
      for (i = 0; i < Nweights; i++)
	Slopes[i] += loss * EtaSlopes[i]
	  * prob_e(Outputs[FirstOutput],j,(j-1));
    }
  } else {
    tmp2 = prob_e(Outputs[FirstOutput], (int)goal[0], (int)goal[1]) / tmp;
    for (i=0; i<Nweights; i++)
      Slopes[i] += -1.0 * EtaSlopes[i] * tmp2;
  }

}

static void vmmin_nolr();

/* transform psi to theta */
static float trans_psi_theta(par, i)
     float par;
     int i;
{
  if (i==1)
     return(SCALEa * par);
  else
     return(exp(SCALEb * par));
}

/* transform theta to psi */
static float trans_theta_psi(par, i)
     float par;
     int i;
{
  if (i==1)
     return(par / SCALEa);
  else
     return(slog(par) / SCALEb);
}

/* calculate dtheta_i / dpsi_i as a function of psi */
static float trans_dtheta_dpsi(par, i)
     float par;
     int i;
{
  if (i==1)
     return(SCALEa);
  else
     return(SCALEb * exp(SCALEb * par));
}

static double fminfn(p)
     float  *p;
{
  int i,j,l,c;
  double sum1,sumgam,theta,tmp,tmp1,detadx,detadx2,st;
  notcomp = false;

  Phi[1] = trans_psi_theta(p[Nweights],1);
  for (i = 2; i < K; i++)
    Phi[i] = Phi[i-1] + trans_psi_theta(p[Nweights+i-1],i);
  Phi[0] = -PHIMAR; Phi[K] = PHIMAR;  
  for (i = 0; i < Nweights; i++)
    wts[i] = p[i];

  TotalError = 0.0;
  sum1 = 0.0;
  quiet = true;
  for (i = 0; i < NTrain; i++) {

    fpass(TrainIn[i], TrainOut[i], Weights[i]);

    if (Mu>0)
      for (l = 1; l < FirstHidden; l++) {
	st = 0.0;
	for (j = FirstHidden; j < FirstOutput; j++)
	  st += wts[wlook[j][FirstOutput]] * 
	    wts[wlook[l][j]] * wts[wlook[l][j]] * 
	    sigmoid_prime_prime(Outputs[j]);
	sum1 += 0.5 * Weights[i] * Mu * st * st;
      }
    
    if ((Gini>0) || (DGini>0)) {

      tmp1 = 0.0;

      for (l = 1; l < FirstHidden; l++) {

	detadx = 0.0 + ((Skip)?(wts[wlook[l][FirstOutput]]):0.0);
	for (j = FirstHidden; j < FirstOutput; j++)
	  detadx += wts[wlook[j][FirstOutput]] * 
	     wts[wlook[l][j]] * sigmoid_prime(Outputs[j]);

	detadx2 = 0.0;
	for (j = FirstHidden; j < FirstOutput; j++)
	  detadx2 += wts[wlook[j][FirstOutput]] * 
	    wts[wlook[l][j]] * wts[wlook[l][j]] *
	    sigmoid_prime_prime(Outputs[j]);
	
	tmp = 0.0;
	for (c = 1; c <= K; c++) {
	  tmp += 2.0 * ( detadx * detadx * 
			 ( prob_e(Outputs[FirstOutput],c,c-1) * 
			   prob_e(Outputs[FirstOutput],c,c-1) + 
			   prob(Outputs[FirstOutput],c,c-1) * 
			   prob_e_e(Outputs[FirstOutput],c,c-1) ) +
			 prob(Outputs[FirstOutput],c,c-1) *
			 prob_e(Outputs[FirstOutput],c,c-1) * detadx2 );
	}
	
	tmp1 += tmp * tmp;

      }
      
      if (Gini>0)
	sum1 += 0.5 * Gini * Weights[i] * tmp1;

      if (DGini>0)
	sum1 += 0.5 * DGini * Weights[i] *
	  (tmp1 - DG[i]) * (tmp1 - DG[i]);

    }

  }
  
  quiet = false;
  if (notcomp) {
    return(0.0);
  }
  
  for (i = 0; i < Nweights; i++) 
    sum1 += Decay[i] * p[i] * p[i];

  sumgam = 0.0;
  if (CutReg)
    for (j = 1; j < (K-1); j++) {
      theta = trans_psi_theta(p[Nweights+j],j+1);
      sumgam += Lambda * theta + (1.0 - Alpha) * slog(theta);
    }
  
  Epoch++;
  return(TotalError + sum1 + sumgam);
}

static void fmingr(p, df)
     float *p, *df;
{
  int c,i,j,k,l,m,n,w,in,out;
  float tmp,tmp1,pp,pe,pee,dodx,eta,pfe,pf,t2;
  
  for (i = 0; i < (K-1); i++)
    ClassSlopes[i] = 0.0;
  Phi[1] = trans_psi_theta(p[Nweights],1);
  for (i = 2; i < K; i++)
    Phi[i] = Phi[i-1] + trans_psi_theta(p[Nweights+i-1],i);
  Phi[0] = -PHIMAR; Phi[K] = PHIMAR;  
  for (i = 0; i < Nweights; i++) {
    wts[i] = p[i];
    Slopes[i] =  2 * Decay[i] * wts[i];
  }

  TotalError = 0.0;

  for (i = 0; i < NTrain; i++) {

    fpass(TrainIn[i], TrainOut[i], Weights[i]);
    bpass(TrainOut[i], Weights[i]);

    if ( (Mu>0) || (Gini>0) || (DGini>0) ) {

      t2 = 0.0;
      for (w = 0; w < Nweights; w++)
	t1weights[w] = 0.0;
      for (c = 0; c < (K-1); c++)
	t1cutpoints[c] = 0.0;

      for (l = 1; l < FirstHidden; l++) {

	for (j = 0; j < Nunits; j++) {
	  doutdx[j] = 0.0;
	  doutdx2[j] = 0.0;
	  delta[j] = 0.0;
	  fac[j] = 0.0;
	}
	
	if (Skip)
	  doutdx[FirstOutput] += wts[wlook[l][FirstOutput]];
	
	for (m = FirstHidden; m < FirstOutput; m++) {
	  doutdx[m] = sigmoid_prime(Outputs[m]) * wts[wlook[l][m]];
	  doutdx2[m] = sigmoid_prime_prime(Outputs[m]) *
	    (wts[wlook[l][m]] * wts[wlook[l][m]]);
	  doutdx[FirstOutput] += doutdx[m] * wts[wlook[m][FirstOutput]];
	  doutdx2[FirstOutput] += doutdx2[m] * wts[wlook[m][FirstOutput]];
	}
	

	for (m = FirstHidden; m < FirstOutput; m++) {
	  fac[m] = (1.0 - 2.0 * Outputs[m]) * doutdx2[m] - 
	    2.0 * doutdx[m] * doutdx[m];
	  delta[m] = wts[wlook[m][FirstOutput]] * doutdx2[FirstOutput];
	}
	
 	for (in = 0; in < Nunits; in++) { 
 	  for (j = Nconn[in]; j < Nconn[in + 1]; j++) {	     
	    out = Conn[j];

	    if (out >= FirstHidden) {
	      dedw[wlook[out][in]] = doutdx2[out] * doutdx2[in];
	    } else if (in < FirstOutput) {

	      dedw[wlook[out][in]] = delta[in] *  
 		(((l==out)?(2.0 * (1.0-2.0*Outputs[in]) * doutdx[in]):0.0) + 
		Outputs[out] * fac[in] );
		
	    } else {
	      dedw[wlook[out][in]] = 0.0;
	    }
	    
	    if (Mu > 0)
	      Slopes[wlook[out][in]] +=
		Mu * Weights[i] * dedw[wlook[out][in]];
	    
	    if ((Gini>0) || (DGini>0)) {

	      if (out >= FirstHidden) {

		detadxdw[wlook[out][in]] = wts[wlook[l][out]] *
		  sigmoid_prime(Outputs[out]);

	      } else if (in < FirstOutput) {

		detadxdw[wlook[out][in]] = wts[wlook[in][FirstOutput]] *
		  ( sigmoid_prime_prime(Outputs[in]) * 
		    wts[wlook[l][in]] * Outputs[out] +
		    ((out==l)?(sigmoid_prime(Outputs[in])):0.0) );
	
	      } else {

		detadxdw[wlook[out][in]] = ((out==l)?1.0:0.0);

	      }

	    }
	    
	  }

	}
	
	if ((Gini>0) || (DGini>0)) {

	  tmp = 0.0;
	  eta = Outputs[FirstOutput];
	  dodx = doutdx[FirstOutput];
	  for (c = 1; c <= K; c++) {
	    tmp += 2.0 * ( dodx * dodx *
			   ( prob_e(eta,c,c-1) * prob_e(eta,c,c-1) + 
			     prob(eta,c,c-1) * prob_e_e(eta,c,c-1) ) +
			   prob(eta,c,c-1) * prob_e(eta,c,c-1) * 
			   doutdx2[FirstOutput] );
	  }
	  
	  t2 += tmp * tmp;


	  for (w = 0; w < Nweights; w++) {
	    tmp1 = 0.0;
	    for (c = 1; c <= K; c++) {
	      pp = prob(eta,c,c-1);
	      pe = prob_e(eta,c,c-1);
	      pee = prob_e_e(eta,c,c-1);
	      
	      tmp1 += 2.0 * (EtaSlopes[w] * dodx * dodx *
			     ( 3.0 * pe * pee + pp * prob_e_e_e(eta,c,c-1) )
			     + 2.0 * dodx * detadxdw[w] * (pe * pe + pp * pee) 
			     + ( pe * pe + pp * pee ) * doutdx2[FirstOutput]
			     * EtaSlopes[w] 
			     + pp * pe * dedw[w] / doutdx2[FirstOutput] );
	      
	    }
	    


	    t1weights[w] += tmp * tmp1;
	    
	  }


	  for (k = 0; k < (K-1); k++) {
	    tmp1 = 0.0;
	    for (c = (k+1); c <= (k+2); c++)
	      {
		pp = prob(eta,c,c-1);
		pe = prob_e(eta,c,c-1);
		pee = prob_e_e(eta,c,c-1);
		pfe = prob_p_e(eta,c,c-1,k+1);
		pf = prob_p(eta,c,c-1,k+1);

		tmp1 += 2.0 * ( dodx * dodx * 
				(2.0 * pe * pfe + pf * pee + 
				 pp * prob_p_e_e(eta,c,c-1,k+1) ) +
				(pf * pe + pp * pfe) * doutdx2[FirstOutput] );

	      }

	    t1cutpoints[k] += tmp * tmp1;
	    
	  }

	}
	
      }
      
      if (Gini>0) {
	for (w = 0; w < Nweights; w++)
	  Slopes[w] += Gini * Weights[i] * t1weights[w];
	
	for (c = 0; c < (K-1); c++)
	  ClassSlopes[c] += Gini * Weights[i] * t1cutpoints[c];
      }
      
      if (DGini>0) {
	for (w = 0; w < Nweights; w++)
	  Slopes[w] += DGini * Weights[i] * 
	    (t2 - DG[i]) * 2.0 * t1weights[w];
	for (c = 0; c < (K-1); c++)
	  ClassSlopes[c] += DGini * Weights[i] * 
	    (t2 - DG[i]) * 2.0 * t1cutpoints[c];
      }
      
    }
    
  }


  for (j = 0; j < Nweights; j++)
    df[j] = Slopes[j];

  for (j = 0; j < (K-1); j++)
    df[Nweights+j] = 0.0;


  for (c=1; c<K; c++)
    for (j=(c-1); j<(K-1); j++)
      df[Nweights+c-1] += ClassSlopes[j];

  if (CutReg)
    for (j = 1; j < (K-1); j++)
      df[Nweights+j] += Lambda + (1.0 - Alpha) / 
	trans_psi_theta(p[Nweights+j],j+1);

  for (j = 0; j < (K-1); j++)
    df[Nweights+j] = df[Nweights+j] * trans_dtheta_dpsi(p[Nweights+j],j+1);

  Epoch++;
}


void Dfminfn(val, p)
     double *val;
     float  *p;
{
  int i;

  for (i = (K-1); i > 1; i--)
    p[Nweights+i-1] -= p[Nweights+i-2];

  for (i = 0; i < (K-1); i++)
    p[Nweights+i] = trans_theta_psi(p[Nweights+i],i+1);

  *val = fminfn(p);

  for (i = 0; i < (K-1); i++)
    p[Nweights+i] = trans_psi_theta(p[Nweights+i],i+1);

  for (i = 2; i < K; i++){ 
    p[Nweights+i-1] += p[Nweights+i-2]; 
  } 
}

void Dfmingr(p, df)
     float  *p,*df;
{
  int i;

  for (i = (K-1); i > 1; i--)
    p[Nweights+i-1] -= p[Nweights+i-2];

  for (i = 0; i < (K-1); i++)
    p[Nweights+i] = trans_theta_psi(p[Nweights+i],i+1);

  fmingr(p,df);

  for (i = 0; i < (K-1); i++)
   df[Nweights+i] = df[Nweights+i] / trans_dtheta_dpsi(p[Nweights+i],i+1);

  for (i = 0; i < (K-2); i++)
    df[Nweights+i] = df[Nweights+i] - df[Nweights+i+1];

  for (i = 0; i < (K-1); i++)
    p[Nweights+i] = trans_psi_theta(p[Nweights+i],i+1);

  for (i = 2; i < K; i++)
    p[Nweights+i-1] += p[Nweights+i-2];

}


void dovm_nolr(Np, params, Fmin, maxit, trace, mask)
     long *Np, *maxit, *trace, *mask;
     float *params;
     double *Fmin;
{
  int i;

  for (i = K-1; i > 1; i--) {
    params[Nweights+i-1] -= params[Nweights+i-2];
  }

  for (i = 0; i < (K-1); i++)
    params[Nweights+i] = trans_theta_psi(params[Nweights+i],i+1);

  vmmin_nolr((int)*Np, params, Fmin, (int)*maxit, (int)*trace, mask);

  for (i = 0; i < (K-1); i++)
    params[Nweights+i] = trans_psi_theta(params[Nweights+i],i+1);

  for (i = 2; i < K; i++){
    params[Nweights+i-1] += params[Nweights+i-2];
  }
}

/* BFGS variable-metric method, based on Pascal code */
/* in J.C. Nash, `Compact Numerical Methods for Computers', 2nd edition, */

static void vmmin_nolr(n0, b,  Fmin, maxit, trace, mask)
     int n0;
     float *b;
     double *Fmin;
     int  maxit;
     int trace;
     long *mask;
{
  boolean accpoint, enough;
  float	  *g, *t, *X, *c, **B;
  int     count, funcount, gradcount;
  double  f, gradproj;
  int     ct, i, j, ilast, iter=0;
  double  s, steplength;
  double  D1, D2;
  int     n, *l;
  double  gtg;

  l = Calloc(n0, int);
  if (!l) errmsg("allocation failure in dovm()");
  n = 0;
  for (i = 0; i < n0; i++)
    if(mask[i]) l[n++] = i;

  g = vect(n0); t = vect(n); X = vect(n); c = vect(n);
  B = Lmatrix(n);

  f = fminfn(b);
  if (notcomp) errmsg("Initial value not computable");
  if(trace) {printf("initial  value %f ",f); fflush(stdout);}
  {
    *Fmin = f;
    funcount = gradcount = 1;
    fmingr(b, g); iter++;
    ilast = gradcount;

    do {
      if (ilast == gradcount) {
	for (i = 0; i < n; i++) {
	  for (j = 0; j < i; j++)
	    B[i][j] = 0.0;
	  B[i][i] = 1.0;
	}
      }
      
      for (i = 0; i < n; i++) {
	X[i] = b[l[i]];
	c[i] = g[l[i]];
      }
      gradproj = 0.0;
      for (i = 0; i < n; i++) {
	s = 0.0;
	for (j = 0; j <= i; j++)
	  s -= B[i][j] * g[l[j]];
	for (j = i+1; j < n; j++)
	  s -= B[j][i] * g[l[j]];
	t[i] = s;
	gradproj += s * g[l[i]];
      }
      
      if (gradproj < 0.0) {
	steplength = STARTSTEP;
	accpoint = false;
	do {
	  count = 0;
	  for (i = 0; i < n; i++) {
	    b[l[i]] = X[i] + steplength * t[i];
	    if (reltest + X[i] == reltest + b[l[i]])
	      count++;
	  }
	  if (count < n) {
	    f = fminfn(b);
	    funcount++;
	    accpoint =
	      (!notcomp) && (f <= *Fmin + gradproj * steplength * acctol);
	    if (!accpoint) steplength *= stepredn;
	  }
	} while (!(count == n || accpoint));

	

  	enough = (f > abstol) && (f < (1.0-reltol)*(*Fmin));

	if(!enough) count = n;
	if (count < n) {
	  *Fmin = f;
	  fmingr(b, g);
	  gradcount++; iter++;
	  D1 = 0.0;
	  for (i = 0; i < n; i++) {
	    t[i] = steplength * t[i];
	    c[i] = g[l[i]] - c[i];
	    D1 += t[i] * c[i];
	  }
	  if (D1 > 0) {
	    D2 = 0.0;
	    for (i = 0; i < n; i++) {
	      s = 0.0;
	      for (j = 0; j <= i; j++)
		s += B[i][j] * c[j];
	      for (j = i+1; j < n; j++)
		s += B[j][i] * c[j];
	      X[i] = s;
	      D2 += s * c[i];
	    }
	    D2 = 1.0 + D2 / D1;
	    for (i = 0; i < n; i++) {
	      for (j = 0; j <= i; j++)
		B[i][j] += (D2 * t[i] * t[j] - X[i] * t[j] - t[i] * X[j]) / D1;
	    }
	  } else {
	    ilast = gradcount;
	  }
	} else {
	  if (ilast < gradcount) {
	    count = 0;
	    ilast = gradcount;
	  }
	}
      } else {	
	count = 0;
	if (ilast == gradcount)
	  count = n;
	else
	  ilast = gradcount;
      }
      if (iter%REPORT == 0 && trace) {
	printf("\niter%4d value %f ",iter,f);
	fflush(stdout);
      }
      if (iter >= maxit) break;
      if (gradcount - ilast > 2*n) {
	ilast = gradcount;
      }
    } while (count != n || ilast != gradcount);
  }
  if(trace) {
    printf("\n final  value %f \n", *Fmin);
    if(iter < maxit) 
      printf("converged\n");
    else 
      printf("stopped after %i iterations\n", iter);
  }

  free_vect(g); free_vect(t); free_vect(X); free_vect(c);
  free_Lmatrix(B,n);
}

static opHessian(input, goal, wx, eta, class, classb)
     float *input, *goal, wx, eta;
     int class, classb;
{
  int i, in1, in2, out1,out2, j, j1,j2, first1, first2;
  double out, s, sum1, sum2, t, tmp, addin, p, d;

  for (i = FirstOutput; i < Nunits; i++) {
    out = Outputs[i];
    s = sigmoid_prime(out);
    t = goal[i - FirstOutput];
    h[i] = 0;
  }
  
  p = prob(eta,class,classb);
  d = prob_e(eta,class,classb);

  for (in1 = 0; in1 < Nunits; in1++)
    for (j1 = Nconn[in1]; j1 < Nconn[in1+1]; j1++) {
      out1 = Conn[j1];
      first1 = (in1 < FirstOutput);
      for (in2 = 0; in2 < Nunits; in2++)
	for (j2 = Nconn[in2]; j2 < Nconn[in2+1]; j2++) {
	  if (j2 <= j1) {
	    addin = 0.0;
	    out2 = Conn[j2];
	    first2 = (in2 < FirstOutput);
	    if ((!first1) && (!first2)) {
	      if(in1 == in2) 
		addin += wx * (h[in1] * Outputs[out1] * Outputs[out2]);
	    } else if (first1 && first2){
	      sum1 = sum2 = 0.0;
	      for(i = FirstOutput; i < Nunits; i++){
		sum1 += Errors[i] * w[in1][i];
		sum2 += w[in1][i] * w[in2][i] * h[i];
	      }
	      s = sigmoid_prime(Outputs[in1]) * 
		sigmoid_prime(Outputs[in2]) * sum2;
	      if(in1 == in2) s += sigmoid_prime_prime(Outputs[in1]) * sum1;
	      addin += wx * (s * Outputs[out1] * Outputs[out2]);
	    } else {
	      if(in1 < in2){
		addin += wx * (Outputs[out1] * sigmoid_prime(Outputs[in1])
			       *(Outputs[out2]*w[in1][in2]*h[in2]
				 +((in1==out2)?Errors[in2]:0)));
	      } else {
		addin += wx * (Outputs[out2] * sigmoid_prime(Outputs[in2])
			       *(Outputs[out1]*w[in2][in1]*h[in1]
				 +((in2==out1)?Errors[in1]:0)));
	      }
	    }	    
	    H[j1][j2] -= d * addin / p;
	  }
	}
    }
}


static float r_p_p(i,s,t)
     int i,s,t;
{
  float f;
  f = 0.0;
  if ( ((s==i)||(s==(i-1))) && ((t==i)||(t==(i-1))) ) {
    f = 1.0 / ((Phi[i] - Phi[i-1]) * (Phi[i] - Phi[i-1]));
    if (s!=t)   /* == */
      f = -1.0 * f;
  }
  return(f);
}


static float l_p_p(eta,class,classb,s,t)
     float eta;
     int class,classb,s,t;
{
  float f,p;
  p = prob(eta,class,classb);
  f = (-prob_p(eta,class,classb,s)*prob_p(eta,class,classb,t)/p 
       + prob_p_p(eta,class,classb,s,t)) / p;
  return(f);
}


static mashHessian(eta,class,classb,wx)
     float eta,wx;
     int class,classb;
{
  int i,j;
  float p,d,a;

  p = prob(eta,class,classb);
  d = prob_e(eta,class,classb);
  a = (prob_e_e(eta,class,classb)-d*d/p)/p;

  for (i=0; i<Nweights; i++)
    for (j=0; j<(i+1); j++)
      H[i][j] -= wx * (EtaSlopes[i]*EtaSlopes[j]*a);
}


static phiHessian(eta,class,classb,wx)
     float eta,wx;
     int class,classb;
{
  int i,j;

  for (i=Nweights; i<(Nweights+K-1); i++)
    for (j=Nweights; j<(i+1); j++) {
      H[i][j] -= wx * l_p_p(eta,class,classb,i-Nweights+1,j-Nweights+1);
    }
}


static mixedHessian(eta,class,classb,wx)
     float eta,wx;
     int class,classb;
{
  int k,i,j;
  float f,g,p,q;
  
  p = prob(eta,class,classb);
  q = prob_e(eta,class,classb);
  
  for (i=Nweights; i<(Nweights+K-1); i++) {
    k = i-Nweights+1;
    g = prob_p(eta,class,classb,k);
    f = wx * (prob_p_e(eta,class,classb,k) - g*q/p) / p;
    for (j=0; j<Nweights; j++)
      H[i][j] -= EtaSlopes[j] * f;
  }
}

void nolrHessian(inwts, phi, Hess)
     float *inwts, *phi, *Hess;
{
  int i, j, s, t, class, classb, row, col;
  float eta, wx, theta;

  if ((Mu>0) || (Gini>0) || (DGini>0))
    errmsg("Analytic evaluation of Hessian only implemented for weight decay and cut-point regularization");

  for (i = 0; i< Nweights; i++)
    wts[i] = inwts[i];
  H = Lmatrix((Nweights+K-1));
  h = vect(Nunits);
  w = matrix(Nunits, Nunits);

  for (i = 0; i < (Nweights+K-1); i++)
    for ( j = 0; j <= i; j++) 
      H[i][j] = 0.0;
  for (j = FirstOutput; j < Nunits; j++)
    for (i = FirstHidden; i < FirstOutput; i++) 
      w[i][j] = 0.0;
  for (j = FirstOutput; j < Nunits; j++) 
    for (i = Nconn[j]; i < Nconn[j+1]; i++) 
      w[Conn[i]][j] = wts[i];
  for (i = 0; i < (K-1); i++)
    ClassSlopes[i] = 0.0;
  for (i = 1; i < K; i++)
    Phi[i] = phi[i-1];
  Phi[0] = -PHIMAR; Phi[K] = PHIMAR;

  for (i = 0; i < NTrain; i++) {
    TotalError = 0.0;
    fpass (TrainIn[i], TrainOut[i], 1.0);
    bpass (TrainOut[i], 1.0);

    eta = Outputs[FirstOutput];
    class = (int)TrainOut[i][0];
    classb = (int)TrainOut[i][1];
    wx = Weights[i];

    opHessian (TrainIn[i], TrainOut[i], wx, eta, class,classb); 
    mashHessian(eta,class,classb,wx);
    mixedHessian(eta,class,classb,wx);
    phiHessian(eta,class,classb,wx);
  }


  if (CutReg)
    for (i=2; i<K; i++)
      for (s=Nweights; s<(Nweights+K-1); s++)
	for (t=Nweights; t<(s+1); t++)
	  H[s][t] += (1.0 - Alpha) * r_p_p(i,s-Nweights+1,t-Nweights+1);
  

  for (i = 0; i < Nweights; i++)
    H[i][i] += 2 * Decay[i];

  for (i = 0; i < (Nweights+K-1); i++)
    for (j = 0; j < (Nweights+K-1); j++)
      *Hess++ = H[max(i,j)][min(i,j)];

  free_vect(h);
  free_matrix(w, Nunits, Nunits);
}


